@@ -7,26 +7,27 @@ from __future__ import print_function
 
 import collections
 from concurrent.futures import Future
-import os
+from getpass import getpass
 import json
+import os
+from pprint import pprint
+import sys
 from threading import Thread, Event, current_thread
 import time
 import types
 import warnings
-from getpass import getpass
-from pprint import pprint
 
 pjoin = os.path.join
 
 import zmq
 from zmq.eventloop.ioloop import IOLoop
 from zmq.eventloop.zmqstream import ZMQStream
-from tornado.gen import multi_future
 
 from traitlets.config.configurable import MultipleInstanceError
 from IPython.core.application import BaseIPythonApplication
 from IPython.core.profiledir import ProfileDir, ProfileDirError
 
+from IPython import get_ipython
 from IPython.utils.capture import RichOutput
 from IPython.utils.coloransi import TermColors
 from jupyter_client.localinterfaces import localhost, is_local_ip
@@ -45,9 +46,10 @@ from ipyparallel import util
 
 from jupyter_client.session import Session
 from ipyparallel import serialize
+from ipyparallel.serialize import PrePickled
 
 from .asyncresult import AsyncResult, AsyncHubResult
-from .futures import MessageFuture
+from .futures import MessageFuture, multi_future
 from .view import DirectView, LoadBalancedView
 
 #--------------------------------------------------------------------------
@@ -99,17 +101,19 @@ class ExecuteReply(RichOutput):
         execute_result = self.metadata['execute_result']
         if execute_result:
             return execute_result.get('data', {})
+        return {}
     
     @property
     def _metadata(self):
         execute_result = self.metadata['execute_result']
         if execute_result:
             return execute_result.get('metadata', {})
-    
+        return {}
+
     def display(self):
         from IPython.display import publish_display_data
         publish_display_data(self.data, self.metadata)
-    
+
     def _repr_mime_(self, mime):
         if mime not in self.data:
             return
@@ -118,52 +122,64 @@ class ExecuteReply(RichOutput):
             return data, self._metadata[mime]
         else:
             return data
-    
+
+    def _repr_mimebundle_(self, *args, **kwargs):
+        data, md = self.data, self.metadata
+        if 'text/plain' in data:
+            data = data.copy()
+            data['text/plain'] = self._plaintext()
+        return data, md
+
     def __getitem__(self, key):
         return self.metadata[key]
-    
+
     def __getattr__(self, key):
         if key not in self.metadata:
             raise AttributeError(key)
         return self.metadata[key]
-    
+
     def __repr__(self):
         execute_result = self.metadata['execute_result'] or {'data':{}}
         text_out = execute_result['data'].get('text/plain', '')
         if len(text_out) > 32:
             text_out = text_out[:29] + '...'
-        
+
         return "<ExecuteReply[%i]: %s>" % (self.execution_count, text_out)
-    
-    def _repr_pretty_(self, p, cycle):
+
+    def _plaintext(self):
         execute_result = self.metadata['execute_result'] or {'data':{}}
         text_out = execute_result['data'].get('text/plain', '')
-        
+
         if not text_out:
-            return
-        
-        try:
-            ip = get_ipython()
-        except NameError:
+            return ''
+
+        ip = get_ipython()
+        if ip is None:
             colors = "NoColor"
         else:
             colors = ip.colors
-        
+
         if colors == "NoColor":
             out = normal = ""
         else:
             out = TermColors.Red
             normal = TermColors.Normal
-        
+
         if '\n' in text_out and not text_out.startswith('\n'):
             # add newline for multiline reprs
             text_out = '\n' + text_out
-        
-        p.text(
-            out + u'Out[%i:%i]: ' % (
+
+        return u''.join([
+            out,
+            u'Out[%i:%i]: ' % (
                 self.metadata['engine_id'], self.execution_count
-            ) + normal + text_out
-        )
+            ),
+            normal,
+            text_out,
+            ])
+
+    def _repr_pretty_(self, p, cycle):
+        p.text(self._plaintext())
 
 
 class Metadata(dict):
@@ -427,18 +443,19 @@ class Client(HasTraits):
         
         if location is not None and addr == localhost():
             # location specified, and connection is expected to be local
-            if not is_local_ip(location) and not sshserver:
+            location_ip = util.ip_for_host(location)
+            if not is_local_ip(location_ip) and not sshserver:
                 # load ssh from JSON *only* if the controller is not on
                 # this machine
                 sshserver=cfg['ssh']
-            if not is_local_ip(location) and not sshserver:
+            if not is_local_ip(location_ip) and not sshserver:
                 # warn if no ssh specified, but SSH is probably needed
                 # This is only a warning, because the most likely cause
                 # is a local Controller on a laptop whose IP is dynamic
                 warnings.warn("""
             Controller appears to be listening on localhost, but not on this machine.
             If this is true, you should specify Client(...,sshserver='you@%s')
-            or instruct your controller to listen on an external IP."""%location,
+            or instruct your controller to listen on an external IP.""" % location,
                 RuntimeWarning)
         elif not sshserver:
             # otherwise sync with cfg
@@ -499,9 +516,8 @@ class Client(HasTraits):
         
         # last step: setup magics, if we are in IPython:
         
-        try:
-            ip = get_ipython()
-        except NameError:
+        ip = get_ipython()
+        if ip is None:
             return
         else:
             if 'px' not in ip.magics_manager.magics:
@@ -642,13 +658,13 @@ class Client(HasTraits):
             self._connected = False
             tb = '\n'.join(content.get('traceback', []))
             raise Exception("Failed to connect! %s" % tb)
-        
+
         self._start_io_thread()
-    
+
     #--------------------------------------------------------------------------
     # handlers and callbacks for incoming messages
     #--------------------------------------------------------------------------
-    
+
     def _unwrap_exception(self, content):
         """unwrap exception, and remap engine_id to int."""
         e = error.unwrap_exception(content)
@@ -825,40 +841,64 @@ class Client(HasTraits):
 
     def _make_io_loop(self):
         """Make my IOLoop. Override with IOLoop.current to return"""
-        return IOLoop()
-    
+        if 'asyncio' in sys.modules:
+            # tornado 5 on asyncio requires creating a new asyncio loop
+            import asyncio
+            try:
+                asyncio.get_event_loop()
+            except RuntimeError:
+                # no asyncio loop, make one
+                # e.g.
+                asyncio.set_event_loop(asyncio.new_event_loop())
+        return IOLoop.current()
+
     def _stop_io_thread(self):
         """Stop my IO thread"""
         if self._io_loop:
             self._io_loop.add_callback(self._io_loop.stop)
         if self._io_thread and self._io_thread is not current_thread():
             self._io_thread.join()
-    
+
+    def _setup_streams(self):
+        self._query_stream = ZMQStream(self._query_socket)
+        self._query_stream.on_recv(self._dispatch_single_reply, copy=False)
+        self._control_stream = ZMQStream(self._control_socket)
+        self._control_stream.on_recv(self._dispatch_single_reply, copy=False)
+        self._mux_stream = ZMQStream(self._mux_socket)
+        self._mux_stream.on_recv(self._dispatch_reply, copy=False)
+        self._task_stream = ZMQStream(self._task_socket)
+        self._task_stream.on_recv(self._dispatch_reply, copy=False)
+        self._iopub_stream = ZMQStream(self._iopub_socket)
+        self._iopub_stream.on_recv(self._dispatch_iopub, copy=False)
+        self._notification_stream = ZMQStream(self._notification_socket)
+        self._notification_stream.on_recv(self._dispatch_notification, copy=False)
+
     def _start_io_thread(self):
         """Start IOLoop in a background thread."""
-        self._io_loop = self._make_io_loop()
-        self._query_stream = ZMQStream(self._query_socket, self._io_loop)
-        self._query_stream.on_recv(self._dispatch_single_reply, copy=False)
-        self._control_stream = ZMQStream(self._control_socket, self._io_loop)
-        self._control_stream.on_recv(self._dispatch_single_reply, copy=False)
-        self._mux_stream = ZMQStream(self._mux_socket, self._io_loop)
-        self._mux_stream.on_recv(self._dispatch_reply, copy=False)
-        self._task_stream = ZMQStream(self._task_socket, self._io_loop)
-        self._task_stream.on_recv(self._dispatch_reply, copy=False)
-        self._iopub_stream = ZMQStream(self._iopub_socket, self._io_loop)
-        self._iopub_stream.on_recv(self._dispatch_iopub, copy=False)
-        self._notification_stream = ZMQStream(self._notification_socket, self._io_loop)
-        self._notification_stream.on_recv(self._dispatch_notification, copy=False)
-        
-        self._io_thread = Thread(target=self._io_main)
+        evt = Event()
+        self._io_thread = Thread(target=self._io_main, args=(evt,))
         self._io_thread.daemon = True
         self._io_thread.start()
-    
-    def _io_main(self):
+        # wait for the IOLoop to start
+        for i in range(10):
+            if evt.wait(1):
+                return
+            if not self._io_thread.is_alive():
+                raise RuntimeError("IO Loop failed to start")
+        else:
+            raise RuntimeError("Start event was never set. Maybe a problem in the IO thread.")
+
+    def _io_main(self, start_evt=None):
         """main loop for background IO thread"""
+        self._io_loop = self._make_io_loop()
+        self._setup_streams()
+        # signal that start has finished
+        # so that the main thread knows that all our attributes are defined
+        if start_evt:
+            start_evt.set()
         self._io_loop.start()
         self._io_loop.close()
-    
+
     @unpack_message
     def _dispatch_single_reply(self, msg):
         """Dispatch single (non-execution) replies"""
@@ -866,7 +906,7 @@ class Client(HasTraits):
         future = self._futures.get(msg_id)
         if future is not None:
             future.set_result(msg)
-    
+
     @unpack_message
     def _dispatch_notification(self, msg):
         """Dispatch notification messages"""
@@ -1170,7 +1210,7 @@ class Client(HasTraits):
         targets = self._build_targets(targets)[0]
         futures = []
         for t in targets:
-            futures.append(self._send(self._control_socket, 'clear_request', content={}, ident=t))
+            futures.append(self._send(self._control_stream, 'clear_request', content={}, ident=t))
         if not block:
             return multi_future(futures)
         for future in futures:
@@ -1213,7 +1253,7 @@ class Client(HasTraits):
         content = dict(msg_ids=msg_ids)
         futures = []
         for t in targets:
-            futures.append(self._send(self._control_socket, 'abort_request',
+            futures.append(self._send(self._control_stream, 'abort_request',
                     content=content, ident=t))
         
         if not block:
@@ -1255,7 +1295,7 @@ class Client(HasTraits):
         
         futures = []
         for t in targets:
-            futures.append(self._send(self._control_socket, 'shutdown_request',
+            futures.append(self._send(self._control_stream, 'shutdown_request',
                         content={'restart':restart},ident=t))
         error = False
         if block or hub:
@@ -1268,7 +1308,7 @@ class Client(HasTraits):
         if hub:
             # don't trigger close on shutdown notification, which will prevent us from receiving the reply
             self._notification_handlers['shutdown_notification'] = lambda msg: None
-            msg = self._send_recv(self._query_socket, 'shutdown_request')
+            msg = self._send_recv(self._query_stream, 'shutdown_request')
             if msg['content']['status'] != 'ok':
                 error = self._unwrap_exception(msg['content'])
             if not error:
@@ -1298,8 +1338,8 @@ class Client(HasTraits):
         Returns
         -------
 
-        executor: distributed.Executor
-            A dask.distributed.Executor connected to the dask cluster.
+        client = distributed.Client
+            A dask.distributed.Client connected to the dask cluster.
         """
         import distributed
 
@@ -1311,7 +1351,7 @@ class Client(HasTraits):
             scheduler_args = dict(scheduler_args) # copy
 
         # Start a Scheduler on the Hub:
-        reply = self._send_recv(self._query_socket, 'become_dask_request',
+        reply = self._send_recv(self._query_stream, 'become_dask_request',
             {'scheduler_args': scheduler_args},
         )
         if reply['content']['status'] != 'ok':
@@ -1326,9 +1366,16 @@ class Client(HasTraits):
         worker_args.setdefault('ncores', 1)
         dview.apply_sync(util.become_dask_worker, **worker_args)
 
-        # Finally, return an Executor connected to the Scheduler
-        executor = distributed.Executor('{ip}:{port}'.format(**distributed_info))
-        return executor
+        # Finally, return a Client connected to the Scheduler
+        try:
+            distributed_Client = distributed.Client
+        except AttributeError:
+            # For distributed pre-1.18.1
+            distributed_Client = distributed.Executor
+
+        client = distributed_Client('{ip}:{port}'.format(**distributed_info))
+
+        return client
 
 
     def stop_dask(self, targets='all'):
@@ -1343,7 +1390,7 @@ class Client(HasTraits):
         dview = self.direct_view(targets)
 
         # Start a Scheduler on the Hub:
-        reply = self._send_recv(self._query_socket, 'stop_distributed_request')
+        reply = self._send_recv(self._query_stream, 'stop_distributed_request')
         if reply['content']['status'] != 'ok':
             raise self._unwrap_exception(reply['content'])
 
@@ -1381,7 +1428,7 @@ class Client(HasTraits):
         metadata = metadata if metadata is not None else {}
 
         # validate arguments
-        if not callable(f) and not isinstance(f, Reference):
+        if not callable(f) and not isinstance(f, (Reference, PrePickled)):
             raise TypeError("f must be callable, not %s"%type(f))
         if not isinstance(args, (tuple, list)):
             raise TypeError("args must be tuple or list, not %s"%type(args))
@@ -1469,7 +1516,7 @@ class Client(HasTraits):
             targets = None
         if targets is not None:
             targets = self._build_targets(targets)[1]
-        return LoadBalancedView(client=self, socket=self._task_socket, targets=targets,
+        return LoadBalancedView(client=self, socket=self._task_stream, targets=targets,
                                 **kwargs)
     
     def executor(self, targets=None):
@@ -1514,7 +1561,7 @@ class Client(HasTraits):
             targets = self._build_targets(targets)[1]
         if single:
             targets = targets[0]
-        return DirectView(client=self, socket=self._mux_socket, targets=targets,
+        return DirectView(client=self, socket=self._mux_stream, targets=targets,
                           **kwargs)
 
     #--------------------------------------------------------------------------
@@ -1603,7 +1650,7 @@ class Client(HasTraits):
         theids = self._msg_ids_from_jobs(indices_or_msg_ids)
         content = dict(msg_ids = theids)
 
-        reply = self._send_recv(self._query_socket, 'resubmit_request', content)
+        reply = self._send_recv(self._query_stream, 'resubmit_request', content)
         content = reply['content']
         if content['status'] != 'ok':
             raise self._unwrap_exception(content)
@@ -1655,7 +1702,7 @@ class Client(HasTraits):
 
         if theids: # some not locally cached
             content = dict(msg_ids=theids, status_only=status_only)
-            reply = self._send_recv(self._query_socket, "result_request", content=content)
+            reply = self._send_recv(self._query_stream, "result_request", content=content)
             content = reply['content']
             if content['status'] != 'ok':
                 raise self._unwrap_exception(content)
@@ -1733,7 +1780,7 @@ class Client(HasTraits):
         else:
             engine_ids = self._build_targets(targets)[1]
         content = dict(targets=engine_ids, verbose=verbose)
-        reply = self._send_recv(self._query_socket, "queue_request", content=content)
+        reply = self._send_recv(self._query_stream, "queue_request", content=content)
         content = reply['content']
         status = content.pop('status')
         if status != 'ok':
@@ -1898,7 +1945,7 @@ class Client(HasTraits):
             msg_ids = self._msg_ids_from_jobs(jobs)
 
         content = dict(engine_ids=targets, msg_ids=msg_ids)
-        reply = self._send_recv(self._query_socket, "purge_request", content=content)
+        reply = self._send_recv(self._query_stream, "purge_request", content=content)
         content = reply['content']
         if content['status'] != 'ok':
             raise self._unwrap_exception(content)
@@ -1954,7 +2001,7 @@ class Client(HasTraits):
                 list of all msg_ids, ordered by task submission time.
         """
 
-        reply = self._send_recv(self._query_socket, "history_request", content={})
+        reply = self._send_recv(self._query_stream, "history_request", content={})
         content = reply['content']
         if content['status'] != 'ok':
             raise self._unwrap_exception(content)
@@ -1978,7 +2025,7 @@ class Client(HasTraits):
         if isinstance(keys, string_types):
             keys = [keys]
         content = dict(query=query, keys=keys)
-        reply = self._send_recv(self._query_socket, "db_request", content=content)
+        reply = self._send_recv(self._query_stream, "db_request", content=content)
         content = reply['content']
         if content['status'] != 'ok':
             raise self._unwrap_exception(content)
